/**
  User-level threads example.
  
  Orion Sky Lawlor, olawlor@acm.org, 2005/2/18 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>
#include <ucontext.h> /* for makecontext/swapcontext routines */
#include <queue> /* C++ STL queue structure */
#include <vector>

/******************* Thread Creation *************/
/* Tiny little user-level thread library */
typedef void (*threadFn)(void);
void addToQueue(ucontext_t *thread);

/* Create a new user-level thread to run this function */
void makeThread(ucontext_t *T,threadFn fn) {
	getcontext(T);
	/* Allocate a stack for the new thread */
	int stackLen=32*1024;
	char *stack=new char[stackLen];
	T->uc_stack.ss_sp=stack;
	T->uc_stack.ss_size=stackLen;
	T->uc_stack.ss_flags=0;
	makecontext(T,fn,0);
	addToQueue(T);
}

/******************* Run Queue *****************/
/* Pool of ready-to-run threads */
std::vector<ucontext_t *> runPool;

/* Add a thread to the pool */
void addToQueue(ucontext_t *thread)
{
	runPool.push_back(thread);
}

/* Get the next ready-to-run thread out of the pool */
ucontext_t *extractFromQueue(void) {
	if (runPool.empty()) {
		printf("Nothing more to run!\n");
		exit(0);
	}
	ucontext_t *next=runPool.back();
	runPool.pop_back();
	return next;
}

/**************** Scheduling *************/
/*
  Suspend this thread for a while, to let other threads run. 
*/
void myThreadYield(ucontext_t *me) {
	addToQueue(me);
	ucontext_t *next=extractFromQueue();
	swapcontext(me,next); /* Switch from running me to running the next guy. */
}

/*
  Just run the next available thread.
*/
void myThreadDone(void) {
	ucontext_t *next=extractFromQueue();
	setcontext(next);
}

/****************** User Code ****************/
/* Context structures for two user-level threads. */
ucontext_t A,B;
int nIter=20;

void runA(void) {
	for (int n=0;n<nIter;n++) {
		for (int i=0;i<1000*1000;i++) {}
		printf("A(%d)\n",n);
		myThreadYield(&A);
	}
	myThreadDone();
}

void runB(void) {
	for (int n=0;n<nIter;n++) {
		for (int i=0;i<1000*1000;i++) {}
		printf("           B(%d)\n",n);
		myThreadYield(&B);
	}
	myThreadDone();
}

int main() {
	makeThread(&A,runA);
	makeThread(&B,runB);
	myThreadDone();
	return 0;
}

